/* main.c
 *   by Alex Chadwick
 * 
 * Copyright (C) 2014, Alex Chadwick
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <bslug.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <rvl/dwc.h>
#include <rvl/GXFrameBuf.h>
#include <rvl/vi.h>

/* support any game id */
BSLUG_MODULE_GAME("????");
BSLUG_MODULE_NAME("GX Console");
BSLUG_MODULE_VERSION("v1.0");
BSLUG_MODULE_AUTHOR("Chadderz");
BSLUG_MODULE_LICENSE("BSD");

/* Replacement for VISetNextFrameBuffer.
 *  notes down the location of all of the games frame buffers so the console can
 *  be displayed to them. Calls the real VISetNextFrameBuffer. */
static void Console_VISetNextFrameBuffer(void *frame_buffer);
/* Replacement for GXSetDispCopyFrame2Field.
 *  sets the copy mode to GX_COPY_INTLC_EVEN which only copies the even scan
 *  lines from the EFB tot he XFB, effectively halving the screen height. Calls
 *  the real GXSetDispCopyFrame2Field with this argument. */
static void Console_GXSetDispCopyFrame2Field(int mode);
/* Replacement for GXSetDispCopyDst.
 *  sets the copy height for the EFB to XFB copy to half the normal value.
 *  Calls the real GXSetDispCopyDst with half the height parameter. */
static void Console_GXSetDispCopyDst(unsigned short width, unsigned short height);
/* Replacement for DWC_SetLogMask.
 *  always enables all logging. Calls DWC_SetLogMask with 0xffffffff. */
static void Console_DWC_SetLogMask(DWC_LogType_t log_mask);
/* Replacement for fwrite.
 *  writes messages to the file 1 and 2 to the screen. File 1 is stdout and file
 *  2 is stderr. All other calls passed through to the real fwrite. */
static size_t Console_fwrite(
    const void *ptr, size_t size, size_t nmemb, FILE *stream);
/* Renders a single character to the screen with specified colour. */
static void Console_PutChar(char c, unsigned int fg);

BSLUG_MUST_REPLACE(VISetNextFrameBuffer, Console_VISetNextFrameBuffer);
BSLUG_MUST_REPLACE(GXSetDispCopyFrame2Field, Console_GXSetDispCopyFrame2Field);
BSLUG_MUST_REPLACE(GXSetDispCopyDst, Console_GXSetDispCopyDst);
/* this method isn't in offline games; don't care if not patched. */
BSLUG_REPLACE(DWC_SetLogMask, Console_DWC_SetLogMask);
BSLUG_MUST_REPLACE(fwrite, Console_fwrite);

/* A bunch of tweakable parameters. Just how tweakable these are really depends.
 * FB_* are related to the game's XFB, SCREEN_* are related to the console pane.
 * Changing the assumption that SCREEN_WIDTH == FB_WIDTH or SCREEN_HEIGHT == 
 * FB_HEIGHT / 2 would cause bad things to happen! */
#define FB_WIDTH console_fb_width
#define FB_HEIGHT console_fb_height
#define FB_DEPTH 2
#define FB_STRIDE ((FB_WIDTH) * (FB_DEPTH))
#define SCREEN_WIDTH FB_WIDTH
#define SCREEN_HEIGHT ((FB_HEIGHT) / 2)
#define SCREEN_STRIDE FB_STRIDE
#define CHAR_WIDTH 8
#define CHAR_HEIGHT 16
#define SCREEN_COLS ((SCREEN_WIDTH) / (CHAR_WIDTH))
#define SCREEN_ROWS ((SCREEN_HEIGHT) / (CHAR_HEIGHT))
#define MAX_FRAME_BUFFER_COUNT 2
#define COLOUR_BG 0x00800080
#define COLOUR_FG 0xFF80FF80
#define COLOUR_FG_STDERR 0x4C544CFF

/* forward declaration of font. */
static const unsigned char console_font[];
/* current row for printing characters */
static int console_cursor_row;
/* current column for printing characters */
static int console_cursor_col;
/* current screen dimensions */
static int console_fb_width;
static int console_fb_height;
/* detected frame buffers */
static void *console_frame_buffers[MAX_FRAME_BUFFER_COUNT];

static void Console_VISetNextFrameBuffer(void *frame_buffer) {
    int i;
    
    /* note down any frame buffers we haven't seen before. */
    for (i = 0; i < MAX_FRAME_BUFFER_COUNT; i++) {
        if (console_frame_buffers[i] == NULL) {
            console_frame_buffers[i] = frame_buffer;
            break;
        } else if (console_frame_buffers[i] == frame_buffer) {
            break;
        }
    }
    /* if we've not seen this frame buffer and we have no room, we reset all
     * memory of frame buffers. This is here in case a game allocates new frame
     * buffers part way in. */
    if (i == MAX_FRAME_BUFFER_COUNT) {
        for (i = 0; i < MAX_FRAME_BUFFER_COUNT; i++) {
            console_frame_buffers[i] = NULL;
        }
        console_frame_buffers[0] = frame_buffer;
    }
    
    /* call down to the real VISetNextFrameBuffer */
    VISetNextFrameBuffer(frame_buffer);
}
static void Console_GXSetDispCopyDst(
    unsigned short width, unsigned short height) {
    console_fb_width = width;
    console_fb_height = height;
    /* call down to the real Console_GXSetDispCopyDst, halving the screen
     * height */
    GXSetDispCopyDst(width, height / 2);
}
static void Console_GXSetDispCopyFrame2Field(int mode) {
    /* call down to the real GXSetDispCopyFrame2Field, replacing the
     * argument. */
    GXSetDispCopyFrame2Field(GX_COPY_INTLC_EVEN);
}
static void Console_DWC_SetLogMask(DWC_LogType_t log_mask) {
    /* call down to the real DWC_SetLogMask enabling all logging. */
    DWC_SetLogMask(0xffffffff);
}
static size_t Console_fwrite(
        const void *ptr, size_t size, size_t nmemb, FILE *stream) {
    if (stream && ptr && (stream->fd == 1 || stream->fd == 2)) {
        /* stdout && stderr */
        const char *str;
        size_t i;
        
        str = ptr;
        /* loop through the string printing each character */
        for (i = 0; i < size * nmemb; i++) {
            Console_PutChar(str[i],
                stream->fd == 1 ? COLOUR_FG : COLOUR_FG_STDERR);
        }
        
        fwrite(ptr, size, nmemb, stream);
        return nmemb;
    } else
        /* call down to original function */
        return fwrite(ptr, size, nmemb, stream);
}
static void Console_PutChar(char c, unsigned int fg) {
    if (FB_WIDTH == 0 || FB_HEIGHT == 0)
        return;
    if (c < ' ') {
        /* special handling for special characters. */
        switch (c) {
        case '\r':
            /* carriage return resets the column. */
            console_cursor_col = 0;
            break;
        case '\n':
            /* line feed moves to the next line. */
            console_cursor_row++;
            console_cursor_col = 0;
            break;
        case '\t':
            /* tab is spaces until we hit a multiple of 4 (at least 1). */
            Console_PutChar(' ', fg);
            while (console_cursor_col & 3)
                Console_PutChar(' ', fg);
            return;
        }
    } else if (c < 128) {
        /* ordinary character handling. */
        int i;
        
        for (i = 0; i < MAX_FRAME_BUFFER_COUNT; i++) {
            char *frame_buffer;
            
            /* must render to all frame buffers. */
            frame_buffer = console_frame_buffers[i];
            
            if (frame_buffer != NULL) {
                int row;
                
                /* skip over game's screen area. */
                frame_buffer += (FB_HEIGHT - SCREEN_HEIGHT) * FB_STRIDE;
                
                for (row = 0; row < CHAR_HEIGHT; row++) {
                    int col;
                    unsigned char val;
                    unsigned int *pp;
                    
                    /* lookup row in font image. */
                    val = console_font[(c - 32) * CHAR_HEIGHT + row];
                    
                    
                    /* annoyingly the frame buffer is stored in YUVU. Thus,
                     * every 4 bytes is 2 pixels. pp is a pointer to the start
                     * of this row of pixels. */
                    pp = (unsigned int *)(frame_buffer +
                        (console_cursor_row * CHAR_HEIGHT + row) * FB_STRIDE +
                        (console_cursor_col * CHAR_WIDTH) * FB_DEPTH);
                    
                    for (col = 0; col < CHAR_WIDTH; col += 2) {
                        /* render the pixel pair */
                        if ((val >> col) & 0x1) {
                            if ((val >> col) & 0x2) {
                                pp[col / 2] = fg;
                            } else {
                                pp[col / 2] = (COLOUR_BG & 0x0000ff00) | (fg & 0xffff00ff);
                            }
                        } else {
                            if ((val >> col) & 0x2) {
                                pp[col / 2] = (COLOUR_BG & 0xffff00ff) | (fg & 0x0000ff00);
                            } else {
                                pp[col / 2] = COLOUR_BG;
                            }                
                        }
                    }
                }
            }
        }
        
        console_cursor_col++;
    }
    
    /* advance the cursor column-wise */
    while (console_cursor_col >= SCREEN_COLS) {
        console_cursor_col -= SCREEN_COLS;
        console_cursor_row++;
    }
    while (console_cursor_col < 0) {
        console_cursor_col += SCREEN_COLS;
        console_cursor_row--;
    }
    /* advance the cursor row-wise */
    while (console_cursor_row >= SCREEN_ROWS) {
        int i;
        
        for (i = 0; i < MAX_FRAME_BUFFER_COUNT; i++) {
            char *frame_buffer;
            
            frame_buffer = console_frame_buffers[i];
            
            if (frame_buffer != NULL) {
                char *scan;
                int row;
                
                scan = frame_buffer + (FB_HEIGHT - SCREEN_HEIGHT) * FB_STRIDE;
                
                /* copy all rows but the first up one row. */
                memmove(
                    scan,
                    scan + FB_STRIDE * CHAR_HEIGHT,
                    FB_STRIDE * (SCREEN_HEIGHT - CHAR_HEIGHT));
                
                scan += FB_STRIDE * (SCREEN_HEIGHT - CHAR_HEIGHT);
                
                /* blank out the last row. */
                for (row = SCREEN_HEIGHT - CHAR_HEIGHT; row < SCREEN_HEIGHT; row++) {
                    int col;
                    unsigned int *scan_u;
                    
                    scan_u = (unsigned int *)scan;
                    
                    for (col = 0; col < SCREEN_WIDTH; col += 2) {
                        *scan_u = COLOUR_BG;
                        scan_u++;
                    }
                    scan += FB_STRIDE;
                }
            }
        }
        
        console_cursor_row--;
    }
}

/* font image. 16 bytes per character starting with ' '. Each byte is monochrome
 * representing a row of pixels such the the lowest bit is the leftmost pixel.
 */
static const unsigned char console_font[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x24, 0x24, 0x24, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x48, 0x24, 0xfe, 0x24, 0x24, 0x7f, 0x12, 0x12, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x52, 0x12, 0x12, 0x3c, 0x50, 0x50, 0x52, 0x3e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x47, 0x49, 0x29, 0x17, 0x10, 0xe8, 0x94, 0x92, 0xe2, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x24, 0x24, 0x1c, 0x4e, 0x4b, 0x31, 0x23, 0xde, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x18, 0x18, 0x18, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x10, 0x10, 0x08, 0x08, 0x0c, 0x04, 0x04, 0x04, 0x0c, 0x08, 0x08, 0x10, 0x10, 0x00, 0x00,
    0x00, 0x08, 0x08, 0x10, 0x10, 0x30, 0x20, 0x20, 0x20, 0x30, 0x10, 0x10, 0x18, 0x08, 0x00, 0x00,
    0x00, 0x00, 0x2c, 0x18, 0x18, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x7e, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x0c, 0x04, 0x04, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x40, 0x20, 0x20, 0x10, 0x10, 0x08, 0x08, 0x04, 0x04, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x66, 0x42, 0x42, 0x5a, 0x42, 0x42, 0x66, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x18, 0x16, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x62, 0x40, 0x60, 0x30, 0x18, 0x0c, 0x06, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x62, 0x40, 0x60, 0x18, 0x60, 0x40, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x30, 0x30, 0x28, 0x24, 0x24, 0x22, 0x7e, 0x20, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x7e, 0x02, 0x02, 0x3a, 0x66, 0x40, 0x40, 0x62, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x38, 0x44, 0x02, 0x3a, 0x66, 0x42, 0x42, 0x64, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x7e, 0x40, 0x20, 0x10, 0x10, 0x18, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x66, 0x42, 0x66, 0x18, 0x66, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x66, 0x42, 0x42, 0x66, 0x5c, 0x40, 0x22, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x18, 0x08, 0x08, 0x04, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x40, 0x30, 0x0c, 0x02, 0x0c, 0x30, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x7e, 0x00, 0x00, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x02, 0x0c, 0x30, 0x40, 0x30, 0x0c, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x62, 0x42, 0x40, 0x20, 0x18, 0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x38, 0x44, 0x42, 0x9a, 0xa5, 0xa5, 0xa5, 0x55, 0x6d, 0x02, 0x42, 0x3c, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x18, 0x18, 0x24, 0x24, 0x24, 0x7e, 0x42, 0x42, 0x81, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3e, 0x62, 0x42, 0x62, 0x3e, 0x42, 0x42, 0x42, 0x3e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x38, 0x66, 0x42, 0x02, 0x02, 0x02, 0x42, 0x44, 0x38, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x1e, 0x62, 0x42, 0x42, 0x42, 0x42, 0x42, 0x62, 0x3e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x7e, 0x02, 0x02, 0x02, 0x7e, 0x02, 0x02, 0x02, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x7e, 0x02, 0x02, 0x02, 0x7e, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x38, 0x46, 0x02, 0x02, 0x72, 0x42, 0x42, 0x46, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x42, 0x42, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x42, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x7e, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x38, 0x20, 0x20, 0x20, 0x20, 0x20, 0x22, 0x26, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x42, 0x22, 0x12, 0x0a, 0x1e, 0x12, 0x22, 0x42, 0xc2, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x7c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x46, 0x66, 0x66, 0x5a, 0x5a, 0x5a, 0x42, 0x42, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x46, 0x46, 0x4a, 0x4a, 0x5a, 0x52, 0x52, 0x62, 0x62, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x66, 0x42, 0x42, 0x42, 0x42, 0x42, 0x66, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3e, 0x42, 0x42, 0x42, 0x62, 0x3e, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x66, 0x42, 0x42, 0x42, 0x42, 0x42, 0x66, 0x3c, 0x10, 0x60, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3e, 0x42, 0x42, 0x42, 0x3e, 0x12, 0x22, 0x62, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x3c, 0x42, 0x02, 0x06, 0x38, 0x40, 0x40, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x7e, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x66, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0xc3, 0x42, 0x42, 0x42, 0x24, 0x24, 0x24, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x81, 0x81, 0xdb, 0x5a, 0x5a, 0x5a, 0x66, 0x66, 0x66, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x42, 0x66, 0x24, 0x18, 0x18, 0x18, 0x24, 0x62, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x42, 0x42, 0x24, 0x24, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x7e, 0x60, 0x20, 0x10, 0x08, 0x0c, 0x04, 0x02, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x38, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x38, 0x00, 0x00,
    0x00, 0x02, 0x04, 0x04, 0x08, 0x08, 0x10, 0x10, 0x20, 0x20, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x1c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1c, 0x00, 0x00,
    0x00, 0x00, 0x18, 0x18, 0x24, 0x24, 0x42, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00,
    0x00, 0x08, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x3c, 0x22, 0x60, 0x7c, 0x62, 0x62, 0xde, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x02, 0x02, 0x02, 0x3a, 0x66, 0x42, 0x42, 0x42, 0x66, 0x3a, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x3c, 0x46, 0x02, 0x02, 0x02, 0x46, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x40, 0x40, 0x40, 0x5c, 0x66, 0x42, 0x42, 0x42, 0x66, 0x5c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x3c, 0x66, 0x42, 0x7e, 0x02, 0x46, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x78, 0x08, 0x08, 0x7e, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x5c, 0x66, 0x42, 0x42, 0x42, 0x66, 0x5c, 0x40, 0x66, 0x3c, 0x00, 0x00,
    0x00, 0x02, 0x02, 0x02, 0x3a, 0x66, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x10, 0x00, 0x00, 0x1c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x10, 0x00, 0x00, 0x1c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0e, 0x00, 0x00,
    0x00, 0x04, 0x04, 0x04, 0x64, 0x34, 0x14, 0x1c, 0x14, 0x24, 0x44, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x1c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x6e, 0x5a, 0x5a, 0x5a, 0x5a, 0x5a, 0x5a, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x3a, 0x66, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x3c, 0x66, 0x42, 0x42, 0x42, 0x66, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x3a, 0x66, 0x42, 0x42, 0x42, 0x66, 0x3a, 0x02, 0x02, 0x02, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x5c, 0x66, 0x42, 0x42, 0x42, 0x66, 0x5c, 0x40, 0x40, 0x40, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x74, 0x0c, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x3c, 0x46, 0x06, 0x38, 0x40, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x04, 0x3e, 0x04, 0x04, 0x04, 0x04, 0x0c, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x42, 0x42, 0x42, 0x42, 0x42, 0x66, 0x5c, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x42, 0x42, 0x66, 0x24, 0x24, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x81, 0xc3, 0x5a, 0x5a, 0x5a, 0x66, 0x66, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x42, 0x24, 0x18, 0x18, 0x1c, 0x24, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x42, 0x42, 0x64, 0x24, 0x2c, 0x18, 0x18, 0x08, 0x08, 0x06, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x7e, 0x20, 0x10, 0x08, 0x0c, 0x06, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x70, 0x18, 0x08, 0x08, 0x08, 0x08, 0x04, 0x08, 0x08, 0x08, 0x08, 0x18, 0x70, 0x00, 0x00,
    0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00,
    0x00, 0x0e, 0x18, 0x10, 0x10, 0x10, 0x10, 0x20, 0x10, 0x10, 0x10, 0x10, 0x18, 0x0e, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x0e, 0x70, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};